/*++++++++++++++
.IDENTIFICATION simbad_client.c
.LANGUAGE       C
.AUTHOR         ESIS / Francois Ochsenbein [CDS]
.ENVIRONMENT    
.KEYWORDS       
.VERSION  1.0   04-Dec-1995 (?)
.VERSION  1.1   13-Jan-2007: Keep previous name
.VERSION  1.2   11-Jan-2009: no local strcpy
.VERSION  1.21  15-Jan-2013: simbad_pos returns -2 if object does not exist
.COMMENTS       
---------------*/
#include <stdio.h>
#include <stdlib.h>		/* getenv(), atof() 			*/
#include <ctype.h>		/* Character Classifications		*/
#include <string.h>		/* strdup	 			*/
#include <simbad_client.h>	/* SIMBAD Cliient Routines Definitions	*/
#include <simcli.h>

/*#define SIMBAD_SERVER "t077esis@simbad.u-strasbg.fr:1674" */
#define SIMBAD_SERVER "v1332viz@simbad.u-strasbg.fr:1674"

static int simbad_plug = -1 ;	/* Connection handle	*/
static char previous_name[80];
static double previous_pos[2];
static int previous_prec = -1;

/*============================================================================*
		Internal Routines
 *============================================================================*/

static int strloc(char *text, int c)
/*++++++++++++++++
.PURPOSE  Locate specified character, preferred to strchr
.RETURNS  Index of located char / length if not found
-----------------*/
{
  char *s;
    for ( s = text; *s && (*s != c) ; s++) ;
    return(s-text);
}

static void init_connection(void)
/*++++++++++++++++
.PURPOSE  Initialize the SIMBAD connection
.RETURNS  The SIMBAD Handle
.REMARKS  The SIMBAD_SERVER environment variable is used for connection
		parameters.
-----------------*/
{
  char *node ;          /* The Simbad-Server Node	*/
  char *service ;       /* The Service number		*/
  char *userid ;        /* SIMBAD User-Name		*/
  char *passwd ;        /* Password for SIMBAD-Server	*/
  char *env, serv1, node1, apasswd[16];
  /* char get_sim_add[80]; -- supprime, FO */
  char *p;

    node = service = userid = passwd = "";
    serv1 = node1 = 0;	/* Saved byte in SIMBAD_SERVER env. variable	*/

    /* Check default, as    USER@NODE:service */
    env = getenv("SIMBAD_SERVER");
    if (!env) env = SIMBAD_SERVER;
    env = strdup(env); /* Non, c'est dangereux: strcpy(get_sim_add,env); */

    if (!*service)	service = env + strloc(env, ':');
    if (*service==':')	serv1 = *service, *(service++) = 0;
    if (!*node)		node =  env + strloc(env, '@');
    if (*node == '@')	node1  = *node, *(node++) = 0;
    if (!*userid)	userid = node1 ? env : "guest";

#if 0	/* ESIS-old */
    if (!*passwd) 	/* Give the password */
	passwd = apasswd,
	passwd[0] = 'e',
	passwd[1] = 's',
	passwd[2] = 'i',
	passwd[3] = 's',
	passwd[4] = '/',
	passwd[5] = 'W',
	passwd[6] = 'e',
	passwd[7] = 'b';
	passwd[8] = 0;
#else
    if (!*passwd) 	/* Give the password */
	passwd = apasswd,
	passwd[0] = 's',
	passwd[1] = 'i',
	passwd[2] = 'm',
	passwd[3] = '4',
	passwd[4] = 'V',
	passwd[5] = 'i',
	passwd[6] = 'z',
	passwd[7] = 0;
	passwd[8] = 0;
#endif

    simbad_plug = simbad_connect(node, service, userid, passwd);
    if (simbad_plug < 0) {
	p = simbad_error(simbad_plug);
	if (!p) p = "(can't get error message)";
	fprintf(stderr, 
	   "****Can't connect to SIMBAD: %s\n    (%s) [%s %s %s %s]\n", 
	   p, env, node, service, userid, passwd) ;
    }
    free(env);
}

static int Fetch(char *astroname)
/*++++++++++++++++
.PURPOSE  Find an Astronomical Object in SIMBAD
.RETURNS  The number of objects found / -1 for error;
.REMARKS  Only the first object is fetched
-----------------*/
{
  int found;

    /* Send the question to SIMBAD */
    found = simbad_query(simbad_plug, astroname, "");
    if (found <= 0)	/* ERROR ***/
    {
	if (found < 0) fprintf(stderr,
	    "****SIMBAD Error for object '%s': %s\n",
		astroname, simbad_error(simbad_plug)) ;
	else fprintf(stderr,
	    "++++SIMBAD Object not found: %s\n", astroname) ;
	return(found) ;
    }	

    /* Retrieve the Object	*/
    if (simbad_retrieve(simbad_plug, 1) < 0) 
    {
	fprintf(stderr, "****SIMBAD Retrieve error: %s\n", 
	    simbad_error(simbad_plug)) ;
	return(0) ;
    }
    return(found);
}

/*============================================================================*
		Public Routines
 *============================================================================*/

int simbad_pos(char *astroname, double opos[2])
/*++++++++++++++++
.PURPOSE  Find the J2000 position of an Astronomical Object
.RETURNS  The precision of the position, as
	-1 (no position) / 0 (degree) / 1 (0.1 deg) / 2 (arcmin) /
	 3 (0.1arcmin) / 4 (arcsec) / 5 (0.1arcsec) / 6 (10mas) / 7 (1mas)
.REMARKS  The search radius should be set to minimal values 3*precision
	(i.e. 3 arcmin when returned precision = 2)
-----------------*/
{
  int found;
  char *s;

    /* V1.2: empty name ==> return -1 */
    if (!astroname) return(-1);
    while (*astroname == ' ') astroname++;
    if (!*astroname) return(-1);

    if (simbad_plug < 0) init_connection();
    if (simbad_plug < 0) return(-1) ;

    /* V1.1: already retrieved ? */
    if (strcmp(astroname, previous_name) == 0) {
	opos[0] = previous_pos[0];
	opos[1] = previous_pos[1];
	return(previous_prec);
    }

    /* Query Simbad, and get the first matching object */
    found = Fetch(astroname);
    if (found <= 0) return(-2) ;	/* V1.21 */

    /* Retrieve the J2000 position	*/
    found = simbad_findata(simbad_plug, "J", "");

    if (found <= 0)	/* No Position 	*/
    {
	fprintf(stderr, "++++SIMBAD Object has no position: %s\n", astroname) ;
	return(-1) ;
    }

    /* The SIMBAD-Position is  RA +DEG (prec)	*/
    s = simbad_getdata(simbad_plug, 1); 
    if (!s)
    {
	fprintf(stderr, "****SIMBAD Error for object '%s': %s\n",
		astroname, simbad_error(simbad_plug)) ;
        return(-1) ;
    }

    /* Decode the RA and Dec */
    while (isspace(*s)) s++;
    opos[0] = atof(s);
    while (isgraph(*s)) s++;
    opos[1] = atof(s);
    s += strloc(s, '(');	/* s -> precision */
    if (*s) s++;
    found = atoi(s)-1;

    /* V1.1: Save the result */
    if (strlen(astroname) < sizeof(previous_name)) {
	strcpy(previous_name, astroname);
	previous_pos[0] = opos[0];
	previous_pos[1] = opos[1];
	previous_prec = found;
    }
    return(found) ;
}

char *simbad_ids(char *astroname)
/*++++++++++++++++
.PURPOSE  Find all aliases fo a given astronomical object
.RETURNS  All aliases separated by commas
-----------------*/
                  	/* IN: Name of an astronomical object	*/
{
  int found;
  char *s, *p, *e;
  static char *list ; static int alist ;

    if (simbad_plug < 0) init_connection();
    if (simbad_plug < 0) return((char *)0) ;

    /* Query Simbad, and get the first matching object */
    found = Fetch(astroname);
    if (found <= 0) return((char *)0) ;
  
    /* Retrieve all aliases	*/
    found = simbad_findata(simbad_plug, "I", "");
    if (found <= 0)	/* No Position 	*/
    {
	if (found < 0) fprintf(stderr,
	    "****SIMBAD_findata for object '%s': %s\n",
		astroname, simbad_error(simbad_plug)) ;
	else fprintf(stderr, 
	    "****SIMBAD_findata has no alias: %s\n", astroname);
	return((char *)0) ;
    }
  
    /* Get now the various aliases */
    if (!list) list = malloc(alist=256) ;
    p = list, e = list + alist - 2 ;
    while (--found >= 0) 
    {
	if ((s = simbad_getdata(simbad_plug, 0))) {
	    if ((p+strlen(s)) >= e) {	/* Must realloc... */
		alist += 256; list = realloc(list, alist) ;
		p = list + strlen(list);
		e = list + alist - 2 ;
	    }
	    if (p != list) *(p++) = ',' ;
	    strcpy(p, s) ;
	    p += strlen(p) ;
	}
	else fprintf(stderr, "++++SIMBAD_getdata for object '%s': %s\n",
	    astroname, simbad_error(simbad_plug)) ;
    }
    return(list) ;
}

int simbad_exit()
/*++++++++++++++++
.PURPOSE  Close all simbad connections
.RETURNS  0=OK
.REMARKS  Nothing done if not opened 
-----------------*/
{
    if (simbad_plug<0) return(0);
    return(simbad_close(simbad_plug));
}

#ifdef TEST
/*============================================================================*
		TEST Program
 *============================================================================*/
int main()
{
  char buf[BUFSIZ] ;
  double pos[2] ;
  int status ; char *ids ;
    while(1) {
	if (isatty(0)) printf("Give a SIMBAD identifier: ") ;
	if (!fgets(buf, sizeof(buf), stdin)) break ;

	if (!isatty(1)) printf("%s: ", buf) ;
	status = simbad_pos(buf, pos) ;
	if (status >= 0) printf("===>Position (%08.4f%+08.4f)", pos[0], pos[1]);
	printf("    status=%d\n", status) ;

	if (!isatty(1)) printf("%s: ", buf) ;
	ids = simbad_ids(buf) ;
	if (ids) printf("===>%s\n", ids) ;
	else     printf("+++>(nil) identifiers ???\n") ;
    }
    simbad_exit();
}
#endif
